/*
 *
 *	    Hex dump and load utility
 *
 *	    John Walker
 *	    formerly of Autodesk Neuchatel
 *	    Avenue des Champs-Montants 14b
 *	    CH-2074 MARIN
 *	    Switzerland
 *	    Usenet: kelvin@Autodesk.com
 *	    Fax:    038/33 88 15
 *	    Voice:  038/33 76 33
 *
 *	Converted to ANSI-C amd ported to MS-DOS by:
 *
 *	Roger Brown, P.E.
 *	29.JUL.1997
 *	r.c.brown@IEEE.org
 */

#define REVDATE "29 JULY 1997"

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define FALSE		0
#define TRUE		1

#define EOS		'\0'

static int		bytesperline = 16, 
			doublechar = FALSE,
			dflen = 2,
			loading = FALSE,
			streaming = FALSE;

static unsigned long	fileaddr;

static unsigned char	lineecho[32];


/************************************************************************/
/*									*/
/*  OUTLINE  --  Edit a line of binary data into the selected output	*/
/*		 format.						*/
/*									*/
/************************************************************************/
static void outline(FILE *out, unsigned char *dat, int len)
{
	static char	addrformat[80] = "%6X",
			dataformat1[80] = "%02X";
	char		oline[132];
	int		i;

	sprintf(oline, addrformat, fileaddr);
	strcat(oline, ":");
	for (i = 0; i < len; i++)
	{
		char outedit[80];

		sprintf(outedit, dataformat1, dat[i]);
		strcat(oline, (i == (bytesperline / 2)) ? "  " : " ");
		strcat(oline, outedit);
	}

	if (doublechar)
	{
		char oc[2];
		int shortfall = ((bytesperline - len) * (dflen + 1)) +
				 (len <= (bytesperline / 2) ? 1 : 0);

		while (shortfall-- > 0)
		{
			strcat(oline, " ");
		}
		oc[1] = EOS;
		strcat(oline, " | ");
		for (i = 0; i < len; i++)
		{
			int b = dat[i];

			/* Map non-printing characters to "." according 
			   to the definitions for ISO 8859/1 Latin-1. */
			if (b < ' ' || (b > '~' && b < 145)
				    || (b > 146 && b < 160))
			{
				b = '.';
			}
			oc[0] = b;
			strcat(oline, oc);
		}
	}
	strcat(oline, "\n");
	fputs(oline, out);
}


/************************************************************************/
/*									*/
/*  INTERPLINE	--  Interpret a line of input.				*/
/*									*/
/************************************************************************/
static int interpline(char *line, int lineno, FILE *out)
{
	char	*cp = line,
		c;
	long	lfaddr;
	int	gfaddr = FALSE;

	/* 
	 * Scan the line for a possible alternative format information
	 * following a vertical bar and delete it.
	 */
	while ((c = *cp++) != EOS)
	{
		if (c == '|')
		{
			cp[-1] = EOS;
			break;
		}
	}

	/*
	 * Scan the line for a file address terminated by a colon.
	 * Save the file address.
	 */
	cp = line;
	while ((c = *cp++) != EOS)
	{
		if (c == ':')
		{
			int sa;
			char tchar;

			cp[-1] = EOS;
			sa = sscanf(line, "%lx%c", &lfaddr, &tchar);
			if (sa == 0 || (sa > 1 && tchar != EOS))
			{
				fprintf(stderr, "xd: Bad hexadecimal file address \"%s\" on line %d:\n", line, lineno);
				return FALSE;
			}
			gfaddr = TRUE;
			cp[-1] = ':';
			break;
		}
	}
	if (!gfaddr)
	{
		cp = line;
	}
	if (!streaming)
	{
		if (!gfaddr)
		{
			fprintf(stderr, "xd: File address missing on line %d:\n", lineno);
			fprintf(stderr, "    %s\n", line);
			return FALSE;
		}
		if (lfaddr != fileaddr)
		{
			fprintf(stderr, "xd: File address sequence error on line %d.\n", lineno);
			fprintf(stderr, "    Expected 0%X, received 0%X.\n", fileaddr, lfaddr);
			fprintf(stderr, "    %s\n", line);
			return FALSE;
		}
	}

	while ((c = *cp++) != EOS)
	{
		if (!isspace(c))
		{
			if (   isxdigit(c)
			    && isxdigit(*cp)
			    && (cp[1] == EOS || isspace(cp[1])))
			{
				int b;
	
				sscanf(cp - 1, "%2x", &b);
				putc(b, out);
				fileaddr++;
				cp += 2;
			}
			else
			{
				fprintf(stderr, "xd: Improper hex value, \"%c%c%c\" on line %d:\n", c, *cp, cp[1], lineno);
				fprintf(stderr, "    %s\n", line);
				fprintf(stderr, "    Bytes must be specified as two hexadecimal\n");
				fprintf(stderr, "    digits separated by white space.\n");
				return FALSE;
			}
		}
	}
	return TRUE;
}


/************************************************************************/
/*									*/
/*  Main program							*/
/*									*/
/************************************************************************/
int main(int argc, char *argv[])
{
	int	i,
		b,
		bp,
		cdata = FALSE,
		f = 0;
	char	*cp,
		*clabel,
		opt;
	FILE	*in = stdin,
		*out = stdout;

	for (i = 1; i < argc; i++)
	{
		cp = argv[i];
		if (*cp == '-')
		{
			opt = *(++cp);
			if (islower(opt))
			{
				opt = toupper(opt);
			}
			switch (opt)
			{
				case 'C':
				doublechar = TRUE;
				break;

				case 'D':
				cdata = TRUE;
				clabel = cp + 1;
				break;

				case 'L':
				loading = TRUE;
				break;

				case 'S':
				streaming = TRUE;
				break;

				case '?':
				case 'H':
				case 'U':
				default :
				fprintf(stdout,"\nXD  --  Hex dump.");
				fprintf(stdout,"\n\nUsage: xd [-option [input [output]]]");
				fprintf(stdout,"\n");
				fprintf(stdout,"\n         Options:");
				fprintf(stdout,"\n              -C      Dump as ISO characters");
				fprintf(stdout,"\n              -Dlabel Dump as a C data declaration");
				fprintf(stdout,"\n              -L      Load file from hex dump");
				fprintf(stdout,"\n              -S      Stream load (don't check file addresses)");
				fprintf(stdout,"\n              -H,U,?  Display this message");
				fprintf(stdout,"\n");
				fprintf(stdout,"\n          (P) Throoput Ltd.  %s", REVDATE);
				fprintf(stdout,"\n                All Rights Reversed");
				fprintf(stdout,"\n");
				return 0;
			}
		}
		else
		{
			switch (f)
			{
				case 0:
#ifdef __unix
				if ((in = fopen(cp, "r")) == NULL)
#else /* !__unix */
				/* Need raw file handling for DOS */
				if ((in = fopen(cp, "rb")) == NULL)
#endif /* __unix */
				{
					fprintf(stderr, "xd: Cannot open input file %s\n", cp);
					return 2;
				}
				f++;
				break;

				case 1:
#ifdef __unix
				if ((out = fopen(cp, "w")) == NULL)
#else /* !__unix */
				/* Need raw file handling for DOS */
				if ((out = fopen(cp, "wb")) == NULL)
#endif /* __unix */
				{
					fprintf(stderr, "xd: Cannot open output file %s\n", cp);
					return 2;
				}
				f++;
				break;

				default:
				fprintf(stderr, "xd: Too many file names specified.\n");
				break;
			}
		}
	}

	bp = 0;
	fileaddr = 0;
	if (loading) 
	{
		char	inline[256];
		int	lineno = 0;

		while (fgets(inline, (sizeof inline) - 2, in)) 
		{
			lineno++;
			if (!interpline(inline, lineno, out)) 
			{
				fclose(out);
				return 2;
			}
		}
	} 
	else 
	{
		if (cdata) 
		{
			char cout[80];

			fprintf(out, "unsigned char %s[] = {\n",clabel[0] == EOS ? "xd_data" : clabel);
			strcpy(cout, "    ");

			while ((b = getc(in)) != EOF) 
			{
				if (strlen(cout) > 72) 
				{
					fprintf(out, "%s\n", cout);
					strcpy(cout, "    ");
				}
				sprintf(cout + strlen(cout), "%d,", b);
			}
			if (strlen(cout) > 4) 
			{
				cout[strlen(cout) - 1] = EOS; /* Strip trailing comma */
				fprintf(out, "%s\n", cout);
			}
			fprintf(out, "};\n");
		} 
		else 
		{
			while ((b = getc(in)) != EOF) 
			{
				if (bp >= bytesperline)
				{
					outline(out, lineecho, bp);
					bp = 0;
					fileaddr += bytesperline;
				}
				lineecho[bp++] = b;
			}

			if (bp > 0)
			{
				outline(out, lineecho, bp);
			}
		}
	}
	return 0;
}
